""" Gecko driver communication handlers """

from .configfile import GeckoConfigFileProtocolHandler
from .firmware import GeckoUpdateFirmwareProtocolHandler
from .getchannel import GeckoGetChannelProtocolHandler
from .hello import GeckoHelloProtocolHandler
from .packcommand import GeckoPackCommandProtocolHandler
from .packet import GeckoPacketProtocolHandler
from .ping import GeckoPingProtocolHandler
from .reminders import GeckoRemindersProtocolHandler
from .statusblock import (
    GeckoStatusBlockProtocolHandler,
    GeckoPartialStatusBlockProtocolHandler,
)
from .version import GeckoVersionProtocolHandler
from .watercare import GeckoWatercareProtocolHandler

__all__ = [
    # From hello
    "GeckoHelloProtocolHandler",
    # From packet
    "GeckoPacketProtocolHandler",
    # From ping
    "GeckoPingProtocolHandler",
    # From version
    "GeckoVersionProtocolHandler",
    # From getchannel
    "GeckoGetChannelProtocolHandler",
    # From configfile
    "GeckoConfigFileProtocolHandler",
    # From statusblock
    "GeckoStatusBlockProtocolHandler",
    "GeckoPartialStatusBlockProtocolHandler",
    # From watercare
    "GeckoWatercareProtocolHandler",
    # From firmware
    "GeckoUpdateFirmwareProtocolHandler",
    # From reminders
    "GeckoRemindersProtocolHandler",
    # From packcommand
    "GeckoPackCommandProtocolHandler",
]



import os
import xml.etree.ElementTree as ET
import logging
import urllib3

from ..const import GeckoConstants

logger = logging.getLogger(__name__)


class GeckoSpaPack:
    """ Class to manage the SpaPackStruct.xml file """

    def __init__(self):
        GeckoSpaPack.download_if_needed()
        self._spa_pack_struct = ET.parse(GeckoConstants.SPA_PACK_STRUCT_FILE)
        self._spa_pack_struct_revision = self._spa_pack_struct.find(
            GeckoConstants.SPA_PACK_REVISION_XPATH
        ).attrib[GeckoConstants.SPA_PACK_REVISION_ATTRIB]
        logger.info("SpaPackStruct v%s", self.revision)

    @property
    def revision(self):
        """ Get the revision of the spa pack structure """
        return self._spa_pack_struct_revision

    @classmethod
    def download_if_needed(cls):
        """ Download SpaPackStruct.xml if it isn't already present """
        if not os.path.exists(GeckoConstants.SPA_PACK_STRUCT_FILE):
            logger.info("SpaPackStruct.xml not found, downloading from the internet...")
            GeckoSpaPack.download()

    @classmethod
    def download(cls):
        """ Download SpaPackStruct.xml from it's permanent home """
        logger.info("Downloading SpaPackStruct.xml")
        http = urllib3.PoolManager()
        response = http.request(
            "GET", GeckoConstants.SPA_PACK_STRUCT_URL, preload_content=False
        )

        with open(GeckoConstants.SPA_PACK_STRUCT_FILE, "wb") as out:
            while True:
                data = response.read(4096)
                if not data:
                    break
                out.write(data)

        response.release_conn()
        logger.info("SpaPackStruct.xml downloaded")

    @property
    def xml(self):
        """ Get the SpaPackStruct.xml root """
        return self._spa_pack_struct



""" GeckoSpa class """

import logging
import threading
import time

from .const import GeckoConstants
from .driver import (
    GeckoUdpSocket,
    GeckoHelloProtocolHandler,
    GeckoPacketProtocolHandler,
    GeckoPingProtocolHandler,
    GeckoVersionProtocolHandler,
    GeckoGetChannelProtocolHandler,
    GeckoConfigFileProtocolHandler,
    GeckoStatusBlockProtocolHandler,
    GeckoPartialStatusBlockProtocolHandler,
    GeckoPackCommandProtocolHandler,
    # Rest
    GeckoSpaPack,
    GeckoStructure,
)
from .automation import GeckoFacade

logger = logging.getLogger(__name__)


class GeckoSpaDescriptor:
    """ A descriptor class for spas that have been discovered on the network """

    def __init__(
        self,
        client_identifier: bytes,
        spa_identifier: bytes,
        spa_name: str,
        sender: tuple,
    ):
        self.client_identifier = client_identifier
        self.identifier = spa_identifier
        self.name = spa_name
        self.ipaddress, self.port = sender

    def get_facade(self, wait_for_connection=True):
        """Get an automation facade to interact with the spa described
        by this class"""
        facade = GeckoFacade(GeckoSpa(self).start_connect())
        if wait_for_connection:
            while not facade.is_connected:
                facade.wait(0.1)
        return facade

    @property
    def identifier_as_string(self):
        return self.identifier.decode(GeckoConstants.MESSAGE_ENCODING)

    @property
    def destination(self):
        return (self.ipaddress, self.port)

    def __repr__(self):
        return f"{self.name}({self.identifier_as_string})"


class GeckoSpa(GeckoUdpSocket, GeckoSpaPack):
    """
    GeckoSpa class manages an instance of a spa, and is the main point of contact for
    control and monitoring. Uses the declarations found in SpaPackStruct.xml to build
    an object that exposes the properties and capabilities of your spa. This class
    should only be used via a Facade which hides the implementation details"""

    def __init__(self, descriptor):
        super().__init__()
        GeckoSpaPack.__init__(self)

        self.descriptor = descriptor
        self.on_connected = None

        self.add_receive_handler(GeckoPacketProtocolHandler())
        self.add_receive_handler(
            GeckoPartialStatusBlockProtocolHandler(
                on_handled=self._on_partial_status_update
            )
        )
        self._ping_thread = threading.Thread(target=self._ping_thread_func, daemon=True)

        # Default values for properties
        self.channel = 0
        self.signal = 0

        self.intouch_version_en = ""
        self.intouch_version_co = ""

        self.gecko_pack_xml = None
        self.config_version = 0
        self.config_xml = None
        self.log_version = 0
        self.log_xml = None
        self.pack_type = None
        self._is_connected = False
        self._connection_started = None
        self.pack = None
        self.version = None
        self.config_number = None
        self.struct = GeckoStructure(self._on_set_value)

    def complete(self):
        """ Complete the use of this spa class """
        super().close()
        self._ping_thread.join()

    @property
    def accessors(self):
        return self.struct.accessors

    def _on_partial_status_update(self, handler, socket, sender):
        for change in handler.changes:
            self.struct.replace_status_block_segment(change[0], change[1])

    def _on_set_value(self, pos, length, newvalue):
        # We issue a pack command to acheive this ...
        self.add_receive_handler(GeckoPackCommandProtocolHandler())
        self.queue_send(
            GeckoPackCommandProtocolHandler.set_value(
                self.get_and_increment_sequence_counter(),
                self.pack_type,
                self.config_version,
                self.log_version,
                pos,
                length,
                newvalue,
                parms=self.sendparms,
            ),
            self.sendparms,
        )

    def _on_config_received(self, handler, socket, sender):
        # XML is case-sensitive, but the platform from the config isn't formed the same,
        # so we manually search the Plateform nodes to find the one we're interested in
        for plateform in self.xml.findall(GeckoConstants.SPA_PACK_PLATEFORM_XPATH):
            if (
                plateform.attrib[GeckoConstants.SPA_PACK_NAME_ATTRIB].lower()
                == handler.plateform_key.lower()
            ):
                self.gecko_pack_xml = plateform

        # We can't carry on without information on how the STATV data block is formed
        if self.gecko_pack_xml is None:
            raise Exception(
                GeckoConstants.EXCEPTION_MESSAGE_NO_SPA_PACK.format(
                    handler.plateform_key
                )
            )

        # Stash the config and log structure declarations
        self.config_version = handler.config_version
        self.config_xml = self.gecko_pack_xml.find(
            GeckoConstants.SPA_PACK_CONFIG_XPATH.format(self.config_version)
        )
        self.log_version = handler.log_version
        self.log_xml = self.gecko_pack_xml.find(
            GeckoConstants.SPA_PACK_LOG_XPATH.format(self.log_version)
        )
        self.pack_type = int(
            self.gecko_pack_xml.attrib[GeckoConstants.SPA_PACK_STRUCT_TYPE_ATTRIB]
        )
        logger.debug(
            "Got spa configuration Type %s - CFG %s/LOG %s, now ask for initial "
            "status block",
            self.pack_type,
            self.config_version,
            self.log_version,
        )

        self.struct.retry_request(
            self,
            GeckoStatusBlockProtocolHandler.full_request(
                socket.get_and_increment_sequence_counter(), parms=sender
            ),
            sender,
        )

    def _on_channel_received(self, handler, socket, sender):
        self.channel = handler.channel
        self.signal = handler.signal_strength
        logger.debug("Got channel %s/%s, now get config", self.channel, self.signal)
        socket.add_receive_handler(
            GeckoConfigFileProtocolHandler(on_handled=self._on_config_received)
        )
        socket.queue_send(
            GeckoConfigFileProtocolHandler.request(
                socket.get_and_increment_sequence_counter(), parms=sender
            ),
            sender,
        )

    def _on_version_received(self, handler, socket, sender):
        self.intouch_version_en = "{0} v{1}.{2}".format(
            handler.en_build, handler.en_major, handler.en_minor
        )
        self.intouch_version_co = "{0} v{1}.{2}".format(
            handler.co_build, handler.co_major, handler.co_minor
        )
        logger.debug(
            "Got software version %s/%s, now get channel",
            self.intouch_version_en,
            self.intouch_version_co,
        )
        socket.add_receive_handler(
            GeckoGetChannelProtocolHandler(on_handled=self._on_channel_received)
        )
        socket.queue_send(
            GeckoGetChannelProtocolHandler.request(
                socket.get_and_increment_sequence_counter(), parms=sender
            ),
            sender,
        )

    @property
    def sendparms(self):
        return (
            self.descriptor.destination[0],
            self.descriptor.destination[1],
            self.descriptor.identifier,
            self.descriptor.client_identifier,
        )

    def _on_ping_response(self, handler, socket, sender):
        self._last_ping = time.monotonic()

    def start_connect(self):
        """ Connect to the in.touch2 module """
        self._connection_started = time.monotonic()
        # Identify self to the intouch module
        self.open()
        self.queue_send(
            GeckoHelloProtocolHandler.client(self.descriptor.client_identifier),
            self.descriptor.destination,
        )
        self._ping_handler = GeckoPingProtocolHandler.request(
            parms=self.sendparms, on_handled=self._on_ping_response
        )
        self.add_receive_handler(self._ping_handler)
        self._ping_thread.start()
        # Get the intouch version
        logger.info("Starting spa connection handshake...")
        self.add_receive_handler(
            GeckoVersionProtocolHandler(on_handled=self._on_version_received)
        )
        self.queue_send(
            GeckoVersionProtocolHandler.request(
                self.get_and_increment_sequence_counter(),
                parms=self.sendparms,
            ),
            self.sendparms,
        )
        return self

    def _loop_func(self):
        if self._is_connected:
            return
        if self.isopen:
            if self.struct.had_at_least_one_block:
                self._final_connect()

    def _final_connect(self):
        logger.debug("Connected, build accessors")
        self.struct.build_accessors([self.config_xml, self.log_xml])
        self.pack = self.accessors[GeckoConstants.KEY_PACK_TYPE].value
        self.version = "{0} v{1}.{2}".format(
            self.accessors[GeckoConstants.KEY_PACK_CONFIG_ID].value,
            self.accessors[GeckoConstants.KEY_PACK_CONFIG_REV].value,
            self.accessors[GeckoConstants.KEY_PACK_CONFIG_REL].value,
        )
        self.config_number = self.accessors[GeckoConstants.KEY_CONFIG_NUMBER].value
        self._is_connected = True
        if self.on_connected is not None:
            self.on_connected(self)
        logger.info("Spa is now connected")

    def _ping_thread_func(self):
        """ Ping thread function """
        logger.info("Ping thread started, %r", self.isopen)
        while self.isopen:
            self.queue_send(self._ping_handler, self.sendparms)
            self.refresh()
            self.wait(GeckoConstants.PING_FREQUENCY_IN_SECONDS)
            if (
                time.monotonic() - self._last_ping
                > GeckoConstants.PING_DEVICE_NOT_RESPONDING_TIMEOUT
            ):
                logger.warning(
                    # TODO
                    "TODO: Spa is not responding to pings, need to reconnect..."
                )
        logger.info("Ping thread finished")

    def get_buttons(self):
        """ Get a list of buttons that can be pressed """
        return [button[0] for button in GeckoConstants.BUTTONS]

    @property
    def is_connected(self):
        if self._is_connected:
            return True
        if (
            time.monotonic() - self._connection_started
            > GeckoConstants.CONNECTION_TIMEOUT_IN_SECONDS
        ):
            raise RuntimeError("Spa took too long to connect ...")
        return False

    def refresh(self):
        """ Refresh the live spa data block """
        if not self.is_connected:
            logger.debug("Can't refresh as we're not connected yet")
            return
        self.struct.retry_request(
            self,
            GeckoStatusBlockProtocolHandler.request(
                self.get_and_increment_sequence_counter(),
                int(self.log_xml.attrib[GeckoConstants.SPA_PACK_STRUCT_BEGIN_ATTRIB]),
                int(self.log_xml.attrib[GeckoConstants.SPA_PACK_STRUCT_END_ATTRIB]),
                parms=self.sendparms,
            ),
            self.sendparms,
        )

    def press(self, keypad):
        """ Simulate a button press """
        self.add_receive_handler(GeckoPackCommandProtocolHandler())
        self.queue_send(
            GeckoPackCommandProtocolHandler.keypress(
                self.get_and_increment_sequence_counter(),
                self.pack_type,
                keypad,
                parms=self.sendparms,
            ),
            self.sendparms,
        )



""" GeckoLib automation interface """

from .base import GeckoAutomationBase
from .blower import GeckoBlower
from .facade import GeckoFacade
from .heater import GeckoWaterHeater
from .keypad import GeckoKeypad
from .light import GeckoLight
from .pump import GeckoPump
from .sensors import GeckoSensor, GeckoBinarySensor
from .switches import GeckoSwitch
from .watercare import GeckoWaterCare

__all__ = [
    "GeckoAutomationBase",
    "GeckoBlower",
    "GeckoFacade",
    "GeckoWaterHeater",
    "GeckoKeypad",
    "GeckoLight",
    "GeckoPump",
    "GeckoSensor",
    "GeckoBinarySensor",
    "GeckoSwitch",
    "GeckoWaterCare",
]



""" Observable class """

import logging

logger = logging.getLogger(__name__)


class Observable:
    """Class to manage observables"""

    def __init__(self):
        self._observers = []

    def watch(self, observer):
        """ Add an observer to this observable class """
        self._observers.append(observer)

    def unwatch(self, observer):
        """ Remove an observer to this observable class """
        self._observers.remove(observer)

    def _on_change(self, sender, old_value, new_value):
        """ Trigger the change notification for all observers """
        logger.info(f"{sender} changed from {old_value} to {new_value}")
        for observer in self._observers:
            observer(sender, old_value, new_value)

    @property
    def has_observers(self):
        return self._observers

    def __repr__(self):
        return f"{self.__class__.__name__} watched by={self._observers!r}"



""" GeckoSimulator class """

import logging
import readline
import os
import glob

from .shared_command import GeckoCmd
from ..driver import (
    GeckoHelloProtocolHandler,
    GeckoPacketProtocolHandler,
    GeckoPingProtocolHandler,
    GeckoVersionProtocolHandler,
    GeckoGetChannelProtocolHandler,
    GeckoConfigFileProtocolHandler,
    GeckoStatusBlockProtocolHandler,
    # GeckoPartialStatusBlockProtocolHandler,
    GeckoWatercareProtocolHandler,
    GeckoUpdateFirmwareProtocolHandler,
    GeckoRemindersProtocolHandler,
    GeckoPackCommandProtocolHandler,
    GeckoStructure,
    GeckoUdpSocket,
)
from .snapshot import GeckoSnapshot
from .. import VERSION

_LOGGER = logging.getLogger(__name__)


class GeckoSimulator(GeckoCmd):
    """GeckoSimulator is a server application to aid with investigating
    Gecko protocol"""

    _STATUS_BLOCK_SEGMENT_SIZE = 39

    def __init__(self, first_commands=None):

        self._socket = GeckoUdpSocket()
        self._install_standard_handlers()
        self.structure = GeckoStructure(self._on_set_value)
        self.snapshot = None

        super().__init__(first_commands)

        self.intro = (
            "Welcome to the Gecko simulator. Type help or ? to list commands.\n"
        )
        self.prompt = "(GeckoSim) "
        readline.set_completer_delims(" \r\n")

    def __exit__(self, *args):
        self._socket.close()
        self._socket = None

    def do_about(self, arg):
        """Display information about this client program and support library : about"""
        del arg
        print("")
        print(
            "GeckoSimulator: A python program using GeckoLib library to simulate Gecko"
            " enabled devices with in.touch2 communication modules"
        )
        print("Library version v{0}".format(VERSION))

    def do_start(self, args):
        """Start the configured simulator : start"""
        if self._socket.isopen:
            print("Simulator is already started")
            return
        self._socket.open()
        self._socket.enable_broadcast()
        self._socket.bind()

    def _complete_path(self, path):
        if os.path.isdir(path):
            return glob.glob(os.path.join(path, "*"))
        return glob.glob(path + "*")

    def do_parse(self, args):
        """Parse logfiles to extract snapshots to the ./snapshot directory. Will
        overwrite identically named snapshot files if present : parse <logfile>"""
        for snapshot in GeckoSnapshot.parse_log_file(args):
            snapshot.save("snapshots")
            print(f"Saved snapshot snapshots/{snapshot.filename}")

    def complete_parse(self, text, line, start_idx, end_idx):
        return self._complete_path(text)

    def do_load(self, args):
        """Load a snapshot : load <snapshot>"""
        snapshots = GeckoSnapshot.parse_log_file(args)
        if len(snapshots) == 1:
            self.set_snapshot(snapshots[0])
            return
        print(
            f"{args} contains {len(snapshots)} snapshots. Please use the"
            f" `parse` command to break it apart"
        )

    def complete_load(self, text, line, start_idx, end_idx):
        return self._complete_path(text)

    def set_snapshot(self, snapshot):
        self.snapshot = snapshot
        self.structure.replace_status_block_segment(0, self.snapshot.bytes)

    def _install_standard_handlers(self):
        """All simulators needs to have some basic functionality such
        as discovery, error handling et al"""
        self._hello_handler = GeckoHelloProtocolHandler.response(
            b"SPA01:02:03:04:05:06", "Udp Test Spa", on_handled=self._on_hello
        )

        self._socket.add_receive_handler(self._hello_handler)
        self._socket.add_receive_handler(GeckoPacketProtocolHandler())
        self._socket.add_receive_handler(
            GeckoPingProtocolHandler(
                on_handled=lambda handler, socket, sender: socket.queue_send(
                    GeckoPingProtocolHandler.response(parms=sender),
                    sender,
                )
            )
        ),
        self._socket.add_receive_handler(
            GeckoVersionProtocolHandler(
                on_handled=lambda handler, socket, sender: socket.queue_send(
                    GeckoVersionProtocolHandler.response(
                        self.snapshot.intouch_EN,
                        self.snapshot.intouch_CO,
                        parms=sender,
                    ),
                    sender,
                )
            )
        )
        self._socket.add_receive_handler(
            GeckoGetChannelProtocolHandler(
                on_handled=lambda handler, socket, sender: socket.queue_send(
                    GeckoGetChannelProtocolHandler.response(10, 33, parms=sender),
                    sender,
                )
            )
        )
        self._socket.add_receive_handler(
            GeckoConfigFileProtocolHandler(
                on_handled=lambda handler, socket, sender: socket.queue_send(
                    GeckoConfigFileProtocolHandler.response(
                        self.snapshot.packtype,
                        self.snapshot.config_version,
                        self.snapshot.log_version,
                        parms=sender,
                    ),
                    sender,
                )
            )
        )
        self._socket.add_receive_handler(
            GeckoStatusBlockProtocolHandler(on_handled=self._on_status_block)
        )
        self._socket.add_receive_handler(
            GeckoWatercareProtocolHandler(on_handled=self._on_watercare)
        )
        self._socket.add_receive_handler(
            GeckoUpdateFirmwareProtocolHandler(
                on_handled=lambda handler, socket, sender: socket.queue_send(
                    GeckoUpdateFirmwareProtocolHandler.response(parms=sender), sender
                )
            )
        )
        self._socket.add_receive_handler(
            GeckoRemindersProtocolHandler(
                on_handled=lambda handler, socket, sender: socket.queue_send(
                    GeckoRemindersProtocolHandler.response(parms=sender), sender
                )
            )
        )
        self._socket.add_receive_handler(
            GeckoPackCommandProtocolHandler(on_handled=self._on_pack_command)
        )

    def _on_hello(self, handler: GeckoHelloProtocolHandler, socket, sender):
        if handler.was_broadcast_discovery:
            socket.queue_send(self._hello_handler, sender)
        elif handler.client_identifier is not None:
            # We're not fussy, we'll chat to anyone!
            pass

    def _on_watercare(self, handler: GeckoWatercareProtocolHandler, socket, sender):
        if handler.schedule:
            socket.queue_send(
                GeckoWatercareProtocolHandler.schedule(parms=sender), sender
            )
        else:
            socket.queue_send(
                GeckoWatercareProtocolHandler.response(1, parms=sender), sender
            )

    def _on_status_block(
        self, handler: GeckoStatusBlockProtocolHandler, socket, sender
    ):
        for idx, start in enumerate(
            range(
                handler.start,
                handler.start + handler.length,
                self._STATUS_BLOCK_SEGMENT_SIZE,
            )
        ):
            length = min(
                self._STATUS_BLOCK_SEGMENT_SIZE,
                len(self.structure.status_block) - start,
            )
            next = (idx + 1) % ((handler.length // self._STATUS_BLOCK_SEGMENT_SIZE) + 1)
            socket.queue_send(
                GeckoStatusBlockProtocolHandler.response(
                    idx,
                    next,
                    self.structure.status_block[start : start + length],
                    parms=sender,
                ),
                sender,
            )

    def _on_pack_command(
        self, handler: GeckoPackCommandProtocolHandler, socket, sender
    ):
        socket.queue_send(
            GeckoPackCommandProtocolHandler.response(parms=sender), sender
        )
        if handler.is_key_press:
            print(f"Pressed a key ({handler.keycode})")
        elif handler.is_set_value:
            print("Set a value")

    def _on_set_value(self, pos, length, newvalue):
        print(f"Simulator: Set value @{pos} for {length} to {newvalue}")



""" Automation switches """

import logging

from .base import GeckoAutomationBase
from .sensors import GeckoSensor

logger = logging.getLogger(__name__)


class GeckoSwitch(GeckoAutomationBase):
    """ A switch can turn something on or off, and can report the current state """

    def __init__(self, facade, key, props):
        """ props is a tuple of (name, keypad_button, state_key, device_class) """
        super().__init__(facade, props[0], key)
        self.ui_key = key
        self._state_sensor = GeckoSensor(
            facade, f"{props[0]} State", self._spa.accessors[props[2]]
        )
        self._state_sensor.watch(self._on_change)
        self._keypad_button = props[1]
        self.device_class = props[3]

    @property
    def is_on(self):
        """ True if the device is ON, False otherwise """
        return self._state_sensor.state != "OFF"

    def turn_on(self):
        """ Turn the device ON, but does nothing if it is already ON """
        logger.debug("%s turn ON", self.name)
        if self.is_on:
            logger.debug("%s request to turn ON ignored, it's already on!", self.name)
            return
        self._spa.press(self._keypad_button)

    def turn_off(self):
        """ Turn the device OFF, but does nothing if it is already OFF """
        logger.debug("%s turn OFF", self.name)
        if not self.is_on:
            logger.debug("%s request to turn OFF ignored, it's already off!", self.name)
            return
        self._spa.press(self._keypad_button)

    def __str__(self):
        return f"{self.name}: {self._state_sensor.state}"



""" Decorator classes """

from ..const import GeckoConstants
from .accessor import GeckoStructAccessor


class GeckoTemperatureDecorator(GeckoStructAccessor):
    """Class to decorate a temperature accessor so that the farenheight tenths, offset
    from freezing are handled"""

    def __init__(self, struct_, accessor):
        super().__init__(struct_, accessor.element)

    def _get_value(self, status_block=None):
        """ Get the temperature """
        # Internally, temp is in farenheight tenths, offset from freezing point
        temp = super()._get_value(status_block)
        units = self.struct.accessors[GeckoConstants.KEY_TEMP_UNITS].value
        if units == "C":
            temp = temp / 18.0
        else:
            temp = (temp + 320) / 10.0
        return temp

    def _set_value(self, temp):
        """ Set the temperature """
        units = self.struct.accessors[GeckoConstants.KEY_TEMP_UNITS].value
        if units == "C":
            temp = float(temp) * 18.0
        else:
            temp = (float(temp) * 10.0) - 320
        super()._set_value(int(temp))



""" Automation interface support classes. """

from ..driver import Observable


class GeckoAutomationBase(Observable):
    """ Base of all the automation helper classes """

    def __init__(self, facade, name, key):
        super().__init__()
        self._facade = facade
        self._spa = facade._spa
        self._name = name
        self._key = key

    @property
    def name(self):
        """ All automation items have a name """
        return self._name

    @property
    def unique_id(self):
        """ A unique id for the property """
        return f"{self._facade.unique_id}-{self._key}"

    @property
    def facade(self):
        """ Return the facade that is associated with this automation object """
        return self._facade



""" Gecko Keypads """

from .base import GeckoAutomationBase


class GeckoKeypad(GeckoAutomationBase):
    """ Keypad management class """

    def __init__(self, facade):
        super().__init__(facade, "Keypad", "KEYPAD")

    def __str__(self):
        return f"{self.name}: Not implemented yet"



""" Spa Structure block """

import logging

from ..const import GeckoConstants
from .protocol import GeckoStatusBlockProtocolHandler
from .udp_socket import GeckoUdpSocket
from .accessor import GeckoStructAccessor
from .decorators import GeckoTemperatureDecorator

logger = logging.getLogger(__name__)


class GeckoStructure:
    """Class to host/manage the raw data block for a spa structure"""

    def __init__(self, on_set_value):
        self._status_block = b"\x00" * 1024
        self.accessors = {}
        self.had_at_least_one_block = False
        self._on_set_value = on_set_value

        # Install a set of handlers

    def replace_status_block_segment(self, offset, segment):
        """ Replace a segment of the status block """
        previous_block = self._status_block
        segment_len = len(segment)
        self._status_block = (
            self._status_block[0:offset]
            + segment
            + self._status_block[offset + segment_len :]
        )
        # Notify changes to accessors
        for accessor in self.accessors.values():
            accessor.status_block_changed(offset, segment_len, previous_block)

    @property
    def status_block(self):
        return self._status_block

    def set_status_block(self, block):
        self._status_block = block

    def _on_retry(self, socket, handler):
        print("Retry")
        if handler is not None:
            handler.reset_timeout()
        self._next_expected = 0
        self._status_block_segments = []

    def _on_status_block_received(
        self, handler: GeckoStatusBlockProtocolHandler, socket, sender
    ):
        if not self._next_expected == handler.sequence:
            logger.warning(
                "Out-of-sequence status block segment %d - ignored", handler.sequence
            )
            if handler.next == 0:
                logger.warning("Retry status block request")
                self._next_expected = 0
                self._status_block_segments = []
                if not handler.retry(socket):
                    raise RuntimeError("Too many retries")
        else:
            self._status_block_segments.append(handler.data)
            self._next_expected = handler.next
            # When we get the last partial segment, we can assume the spa is
            # connected and we can report on status
            if handler.next == 0:
                logger.info("Status block segments complete, update and remove handler")
                self.replace_status_block_segment(
                    self._status_block_offset, b"".join(self._status_block_segments)
                )
                self.had_at_least_one_block = True
                handler._should_remove_handler = True

    def build_accessors(self, xmllist):
        self.accessors = {
            element.tag: GeckoStructAccessor(self, element)
            for xml in xmllist
            for element in xml.findall(
                GeckoConstants.SPA_PACK_STRUCT_POS_ELEMENTS_XPATH
            )
        }
        # Fix temperature accessors ...
        temp_keys = {
            element.tag
            for xml in xmllist
            for element in xml.findall(
                GeckoConstants.SPA_PACK_STRUCT_WORD_TYPE_ELEMENTS_XPATH
            )
            if "temp" in element.tag.lower() or "setpoint" in element.tag.lower()
        }
        logger.debug("Temperature keys to decorate %s", temp_keys)
        for key in temp_keys:
            self.accessors[key] = GeckoTemperatureDecorator(self, self.accessors[key])

    def retry_request(
        self,
        socket_: GeckoUdpSocket,
        request: GeckoStatusBlockProtocolHandler,
        sender: tuple,
    ):
        request._on_handled = self._on_status_block_received
        self._status_block_offset = request.start
        socket_.add_receive_handler(request)
        self._next_expected = 0
        self._status_block_segments = []
        socket_.queue_send(request, sender)

    def set_value(self, pos, length, newvalue):
        # Delegate this
        self._on_set_value(pos, length, newvalue)



""" Gecko <HELLO> handlers """

import logging
from ..udp_socket import GeckoUdpProtocolHandler
from ...const import GeckoConstants

HELLO_OPEN = b"<HELLO>"
HELLO_CLOSE = b"</HELLO>"

_LOGGER = logging.getLogger(__name__)


class GeckoHelloProtocolHandler(GeckoUdpProtocolHandler):
    def __init__(self, content: bytes, **kwargs):
        super().__init__(**kwargs)
        self._send_bytes = b"".join([HELLO_OPEN, content, HELLO_CLOSE])
        self.was_broadcast_discovery = False
        self.client_identifier = None
        self.spa_identifier = None
        self.spa_name = None

    @staticmethod
    def broadcast_address():
        return (GeckoConstants.BROADCAST_ADDRESS, GeckoConstants.INTOUCH2_PORT)

    @staticmethod
    def broadcast(**kwargs):
        return GeckoHelloProtocolHandler(b"1", **kwargs)

    @staticmethod
    def client(client_identifier: bytes, **kwargs):
        return GeckoHelloProtocolHandler(client_identifier, **kwargs)

    @staticmethod
    def response(spa_identifier: bytes, spa_name: str, **kwargs):
        return GeckoHelloProtocolHandler(
            b"".join(
                [spa_identifier, b"|", spa_name.encode(GeckoConstants.MESSAGE_ENCODING)]
            ),
            **kwargs,
        )

    def can_handle(self, received_bytes: bytes, sender: tuple) -> bool:
        return received_bytes.startswith(HELLO_OPEN) and received_bytes.endswith(
            HELLO_CLOSE
        )

    def handle(self, socket, received_bytes: bytes, sender: tuple):
        content = received_bytes[7:-8]
        self.was_broadcast_discovery = False
        self.client_identifier = self.spa_identifier = self.spa_name = None
        if content == b"1":
            self.was_broadcast_discovery = True
        elif content.startswith(b"IOS") or content.startswith(b"AND"):
            self.client_identifier = content
        else:
            self.spa_identifier, spa_name = content.split(b"|")
            self.spa_name = spa_name.decode(GeckoConstants.MESSAGE_ENCODING)

    def __repr__(self):
        return (
            f"{super().__repr__()} (was_broadcast={self.was_broadcast_discovery},"
            f" client_id={self.client_identifier}, spa_id={self.spa_identifier},"
            f" spa_name={self.spa_name} )"
        )



""" Gecko driver """

from .decorators import GeckoTemperatureDecorator
from .protocol import (
    GeckoHelloProtocolHandler,
    GeckoPacketProtocolHandler,
    GeckoPingProtocolHandler,
    GeckoVersionProtocolHandler,
    GeckoGetChannelProtocolHandler,
    GeckoConfigFileProtocolHandler,
    GeckoStatusBlockProtocolHandler,
    GeckoPartialStatusBlockProtocolHandler,
    GeckoWatercareProtocolHandler,
    GeckoUpdateFirmwareProtocolHandler,
    GeckoRemindersProtocolHandler,
    GeckoPackCommandProtocolHandler,
)
from .observable import Observable

from .spapack import GeckoSpaPack
from .accessor import GeckoStructAccessor
from .spastruct import GeckoStructure
from .udp_socket import GeckoUdpProtocolHandler, GeckoUdpSocket

__all__ = [
    "GeckoTemperatureDecorator",
    #
    "GeckoHelloProtocolHandler",
    "GeckoPacketProtocolHandler",
    "GeckoPingProtocolHandler",
    "GeckoVersionProtocolHandler",
    "GeckoGetChannelProtocolHandler",
    "GeckoConfigFileProtocolHandler",
    "GeckoStatusBlockProtocolHandler",
    "GeckoPartialStatusBlockProtocolHandler",
    "GeckoWatercareProtocolHandler",
    "GeckoUpdateFirmwareProtocolHandler",
    "GeckoRemindersProtocolHandler",
    "GeckoPackCommandProtocolHandler",
    "Observable",
    #
    "GeckoSpaPack",
    "GeckoStructure",
    "GeckoStructAccessor",
    "GeckoUdpProtocolHandler",
    "GeckoUdpSocket",
]



""" Structure accessor """

import struct
import logging

from ..const import GeckoConstants
from .observable import Observable

logger = logging.getLogger(__name__)


class GeckoStructAccessor(Observable):
    """Class to access the spa data structure according to the declaration
    in SpaPackStruct.xml"""

    def __init__(self, struct_, element):
        super().__init__()
        self.element = element
        self.tag = element.tag
        self.struct = struct_
        self.pos = int(element.attrib[GeckoConstants.SPA_PACK_STRUCT_POS_ATTRIB])
        self.type = element.attrib[GeckoConstants.SPA_PACK_STRUCT_TYPE_ATTRIB]
        self.bitpos = None

        if GeckoConstants.SPA_PACK_STRUCT_BITPOS_ATTRIB in element.attrib:
            self.bitpos = int(
                element.attrib[GeckoConstants.SPA_PACK_STRUCT_BITPOS_ATTRIB]
            )
            self.bitmask = 1
        if GeckoConstants.SPA_PACK_STRUCT_ITEMS_ATTRIB in element.attrib:
            self.items = element.attrib[
                GeckoConstants.SPA_PACK_STRUCT_ITEMS_ATTRIB
            ].split("|")

        self.length = 1
        self.format = ">B"

        if GeckoConstants.SPA_PACK_STRUCT_SIZE_ATTRIB in element.attrib:
            self.length = int(
                element.attrib[GeckoConstants.SPA_PACK_STRUCT_SIZE_ATTRIB]
            )
            if self.length == 2:
                self.format = ">H"

        if (
            self.type == GeckoConstants.SPA_PACK_STRUCT_WORD_TYPE
            or self.type == GeckoConstants.SPA_PACK_STRUCT_TIME_TYPE
        ):
            self.length = 2
            self.format = ">H"
        if GeckoConstants.SPA_PACK_STRUCT_MAXITEMS_ATTRIB in element.attrib:
            self.maxitems = int(
                element.attrib[GeckoConstants.SPA_PACK_STRUCT_MAXITEMS_ATTRIB]
            )
            if self.maxitems > 8:
                self.bitmask = 15
            elif self.maxitems > 4:
                self.bitmask = 7
            elif self.maxitems > 2:
                self.bitmask = 3
        self.read_write = None
        if GeckoConstants.SPA_PACK_STRUCT_READ_WRITE_ATTRIB in element.attrib:
            self.read_write = element.attrib[
                GeckoConstants.SPA_PACK_STRUCT_READ_WRITE_ATTRIB
            ]

    def status_block_changed(self, offset, len, previous):
        # Does the notified range intersect us, if not then we don't care!
        intersection_start = max(offset, self.pos)
        intersection_end = min(offset + len, self.pos + self.length)
        if intersection_end - intersection_start <= 0:
            return

        logger.debug(
            "Accessor %s @ %d-%d notified of change at %d-%d",
            self.tag,
            self.pos,
            self.pos + self.length,
            offset,
            offset + len,
        )

        old_value = self._get_value(previous)
        new_value = self.value

        # No reason to notify if there is no change
        if new_value == old_value:
            return

        logger.info(
            "Value for %s changed from %s to %s", self.tag, old_value, new_value
        )

        self._on_change(self, old_value, new_value)

    def _get_value(self, status_block=None):
        """Get a value from the pack structure using the initialized declaration
        or using the optionally supplied status_block (used by change notification)"""
        if status_block is None:
            status_block = self.struct.status_block
        data = struct.unpack(
            self.format, status_block[self.pos : self.pos + self.length]
        )[0]
        logger.debug(
            "Accessor %s @ %s, %s raw data = %x", self.tag, self.pos, self.type, data
        )
        if self.bitpos is not None:
            data = (data >> self.bitpos) & self.bitmask
            logger.debug(
                "BitPos %s accessor %s adjusted data = %x",
                (self.bitpos, self.bitmask),
                self.tag,
                data,
            )
        if self.type == GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE:
            data = data == 1
            logger.debug("Bool accessor %s adjusted data = %s", self.tag, data)
        elif self.type == GeckoConstants.SPA_PACK_STRUCT_ENUM_TYPE:
            try:
                data = self.items[data]
                logger.debug("Enum accessor %s adjusted data = %s", self.tag, data)
            except IndexError:
                logger.exception(
                    "Enum accessor %s out-of-range for %s", self.tag, self.items
                )
        return data

    def _set_value(self, newvalue):
        """ Set a value in the pack structure using the initialized declaration """
        if self.read_write is None:
            raise Exception(
                GeckoConstants.EXCEPTION_MESSAGE_NOT_WRITABLE.format(self.tag)
            )

        if self.type == GeckoConstants.SPA_PACK_STRUCT_ENUM_TYPE:
            logger.debug("Enum accessor %s adjusted from %s", self.tag, newvalue)
            newvalue = self.items.index(newvalue)

        # If it is a bitpos, then mask it with the existing value
        existing = struct.unpack(
            self.format, self.struct.status_block[self.pos : self.pos + self.length]
        )[0]
        if self.bitpos is not None:
            logger.debug(
                "Bitpos %s accessor %s adjusted from %s",
                (self.bitpos, self.bitmask),
                self.tag,
                newvalue,
            )
            newvalue = (existing & ~(self.bitmask << self.bitpos)) | (
                (newvalue & self.bitmask) << self.bitpos
            )

        logger.debug(
            "Accessor %s @ %s, %s setting value to %s, existing value was %s. "
            "Length is %d",
            self.tag,
            self.pos,
            self.type,
            newvalue,
            existing,
            self.length,
        )

        # We can't handle this here, we must delegate via the structure
        self.struct.set_value(self.pos, self.length, newvalue)

    @property
    def value(self):
        """ Get a value from the pack structure using the initialized declaration """
        return self._get_value()

    @value.setter
    def value(self, newvalue):
        """ Set a value in the pack structure using the initialized declaration """
        self._set_value(newvalue)

    def __repr__(self):
        return f"{self.tag!r}: {self.value!r}"



""" GeckoUdpSocket - Gecko UDP socket implementation """

import socket
import logging
import threading
import time

_LOGGER = logging.getLogger(__name__)


class GeckoUdpProtocolHandler:
    """
    Protocol handlers manage both sides of a specific conversation part with
    a remote end.

    The protocol is either initiated by a client or by a server, but either
    way a query should always met with a response from the remote end

    Both sides may instantiate listening handlers which will deal with
    unsolicited requests from remote clients and will respond so that the
    remote end knows the request was received and they may also send a query
    to the remote end, expecting a response to confirm receipt.

    A message sent will either make it to the destination, or it won't. Since
    this protocol is built on UDP, we have to cook our own timeout/retry
    mechanism to ensure delivery ... oddly this is precisely what TCP already
    does, no idea why this wasn't considered but hey-ho, this is all a bit of
    fun anyway!

    The base protocol handler will manage the lifetime of the handlers within
    the socket, and mechanisms are in place to allow retries to be handled
    with failure exit points available to allow clients to make class decisions
    or instance decisions, either by overridden methods, or by instance handlers

    """

    def __init__(self, **kwargs):
        # Send functionality
        self._send_bytes = kwargs.get("send_bytes", None)
        self.last_destination = None

        # Receive functionality
        self._on_handled = kwargs.get("on_handled", None)

        # Lifetime functionality
        self._start_time = time.monotonic()
        self._timeout_in_seconds = kwargs.get("timeout", 0)
        self._retry_count = kwargs.get("retry_count", 0)
        self._should_remove_handler = False

    ##########################################################################
    #
    #                         SEND FUNCTIONALITY
    #
    ##########################################################################
    @property
    def send_bytes(self) -> bytes:
        """The bytes to send to the remote end. Either uses the class instance
        data _send_bytes or can be overridden in a base class"""
        if self._send_bytes is None:
            raise NotImplementedError
        return self._send_bytes

    ##########################################################################
    #
    #                        RECEIVE FUNCTIONALITY
    #
    ##########################################################################

    def can_handle(self, received_bytes: bytes, sender: tuple) -> bool:
        """Check if you can handle these bytes. If you return True, then your
        handle method will get called and no other handlers will be given a
        chance to process this data. If you return False then the search for a
        suitable handler will continue"""
        return False

    def handle(self, socket, received_bytes: bytes, sender: tuple):
        """Handle this data. This will only be called if you returned True
        from the `can_handle` function. If you wish to remove this handler
        from the system, then you should set the `should_remove_handler`
        member."""

    def handled(self, socket, sender: tuple):
        self._reset_timeout()
        if self._on_handled is not None:
            self._on_handled(self, socket, sender)

    ##########################################################################
    #
    #                         LIFETIME MANAGEMENT
    #
    ##########################################################################
    @property
    def age(self):
        return time.monotonic() - self._start_time

    @property
    def has_timedout(self):
        return (
            self.age > self._timeout_in_seconds
            if self._timeout_in_seconds > 0
            else False
        )

    @property
    def should_remove_handler(self):
        return self._should_remove_handler

    def _reset_timeout(self):
        self._start_time = time.monotonic()

    def retry(self, socket):
        if self._retry_count == 0:
            return False
        self._retry_count -= 1
        _LOGGER.debug("Handler retry count %d", self._retry_count)
        self._reset_timeout()
        if socket is not None:
            # Queue another send
            socket.queue_send(self, self.last_destination)
        return True

    def loop(self, socket):
        """ Executed each time around the socket loop """
        if not self.has_timedout:
            return
        _LOGGER.debug("Handler has timed out")
        if self.retry(socket):
            return
        self._should_remove_handler = True

    # Pythonic methods
    def __repr__(self):
        return (
            f"{self.__class__.__name__}(send_bytes={self._send_bytes!r},"
            f" age={self.age}, has_timedout={self.has_timedout},"
            f" should_remove_handler={self.should_remove_handler},"
            f" timeout={self._timeout_in_seconds}s,"
            f" retry_count={self._retry_count}"
            f")"
        )


class GeckoUdpSocket:
    """Gecko in.touch2 uses UDP to communicate. This class is a wrapper around
    a socket and a thread and serviced by classes derived from GeckoUdpSendHandler
    GeckoUdpReceiveHandler and GeckoUdpProtocolHandler
    """

    _PORT = 10022
    _SOCKET_TIMEOUT = 0.05
    _MAX_PACKET_SIZE = 8192
    _SENDING_THROTTLE_RATE_PER_SECOND = 50

    def __init__(self, socket=None):
        self._socket = socket
        self._exit_event = None
        self._thread = None
        self._lock = threading.Lock()
        self._send_handlers = []
        self._receive_handlers = []
        self._busy_count = 0
        self._sequence_counter = 0
        self._last_send_time = time.monotonic()

    def __enter__(self):
        if self._socket is None:
            self._socket = socket.socket(
                socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
            )
        self._socket.settimeout(self._SOCKET_TIMEOUT)
        self._exit_event = threading.Event()
        self._thread = threading.Thread(target=self._thread_func, daemon=True)
        self._thread.start()
        return self

    def __exit__(self, *args):
        if self._exit_event:
            self._exit_event.set()
        if self._thread:
            self._thread.join()
            self._thread = None
        if self._socket:
            self._socket.close()
            self._socket = None

    class _BusyLock:
        def __init__(self, socket):
            self._socket = socket

        def __enter__(self):
            with self._socket._lock:
                self._socket._busy_count += 1

        def __exit__(self, *args):
            with self._socket._lock:
                self._socket._busy_count -= 1

    def open(self):
        """Start the use of this UDP socket object if not used
        in a `with` statement"""
        self.__enter__()

    def close(self):
        """Finish the use of this UDP socket object if not used
        in a `with` statement"""
        self.__exit__()

    @property
    def isopen(self):
        """Check to see if the socket is open"""
        if not self._exit_event:
            return False
        return not self._exit_event.is_set()

    @property
    def isbusy(self):
        """Check to see if the socket is busy"""
        if self._send_handlers:
            return True
        if self._receive_handlers:
            return True
        return self._busy_count > 0

    def wait(self, timeout):
        """ Wait for a timeout, respecting the exit event """
        self._exit_event.wait(timeout)

    def bind(self):
        """Bind this UDP socket to the local address and port"""
        self._socket.bind(("", self._PORT))

    def enable_broadcast(self):
        """Set this UDP socket to support broadcast"""
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    def add_receive_handler(self, protocol_handler: GeckoUdpProtocolHandler):
        """Add a receive handler to the list of available handlers"""
        with self._lock:
            self._receive_handlers.append(protocol_handler)

    def remove_receive_handler(self, protocol_handler: GeckoUdpProtocolHandler):
        """Remove a receive handler from the list of available handlers"""
        with self._lock:
            self._receive_handlers.remove(protocol_handler)

    def queue_send(self, protocol_handler: GeckoUdpProtocolHandler, destination: tuple):
        """Queue a message to be sent by the worker thread"""
        with self._lock:
            self._send_handlers.append((protocol_handler, destination))

    def get_and_increment_sequence_counter(self):
        with self._lock:
            self._sequence_counter += 1
            return self._sequence_counter % 256

    def _process_send_requests(self):
        # Throttle the sending rate to prevent message loss
        if (time.monotonic() - self._last_send_time) < (
            1.0 / self._SENDING_THROTTLE_RATE_PER_SECOND
        ):
            return

        with GeckoUdpSocket._BusyLock(self):
            # Assume there are no send requests
            send_handler = None
            # Play safely with clients, and minimize the time spend locked
            with self._lock:
                if self._send_handlers:
                    send_handler = self._send_handlers.pop(0)
            # If there is a handler, then use it
            if send_handler:
                try:
                    send_bytes = send_handler[0].send_bytes
                    destination = send_handler[1]
                    if destination is None:
                        raise AssertionError(
                            f"Cannot have destination set to None for {send_handler}"
                        )
                    # For convenience, the entire destination sometimes gets passed in,
                    # this fixes it to be just the address and port
                    if len(destination) > 2:
                        destination = (destination[0], destination[1])
                    send_handler[0].last_destination = destination
                    _LOGGER.debug("Sending %s to %s", send_bytes, destination)
                    self._socket.sendto(send_bytes, destination)
                    self._last_send_time = time.monotonic()
                except Exception:
                    _LOGGER.exception("Exception during send processing")

    def dispatch_recevied_data(self, received_bytes: bytes, remote_end: tuple):
        """ Dispatch bytes to the handlers, maybe someone is interested! """
        with GeckoUdpSocket._BusyLock(self):
            _LOGGER.debug("Received %s from %s", received_bytes, remote_end)
            receive_handler = None
            with self._lock:
                for handler in self._receive_handlers:
                    if handler.can_handle(received_bytes, remote_end):
                        receive_handler = handler
                        break

            if receive_handler:
                try:
                    receive_handler.handle(self, received_bytes, remote_end)
                    receive_handler.handled(self, remote_end)
                except Exception:
                    _LOGGER.exception("Unhandled exception in receive_handler func")
            else:
                _LOGGER.warning("Couldn't find new handler for %s", received_bytes)

    def _process_received_data(self):
        with GeckoUdpSocket._BusyLock(self):
            try:
                received_bytes, remote_end = self._socket.recvfrom(
                    self._MAX_PACKET_SIZE
                )
                self.dispatch_recevied_data(received_bytes, remote_end)
            except (socket.timeout, OSError):
                return
            finally:
                pass

    def _cleanup_handlers(self):
        with GeckoUdpSocket._BusyLock(self):
            remove_handlers = []

            with self._lock:
                # Build list of handlers that need to be removed
                remove_handlers = [
                    handler
                    for handler in self._receive_handlers
                    if handler.should_remove_handler
                ]

            if remove_handlers:
                _LOGGER.debug("Removed timedout handlers %s", remove_handlers)

            # Remove them from the collection
            with self._lock:
                self._receive_handlers = [
                    handler
                    for handler in self._receive_handlers
                    if handler not in remove_handlers
                ]

            if remove_handlers:
                _LOGGER.debug("Remaining handlers %s", self._receive_handlers)

    def _thread_func(self):
        while self.isopen:
            self._process_send_requests()
            self._process_received_data()
            # Do loop for timeout/retry
            for handler in self._receive_handlers:
                handler.loop(self)
            self._cleanup_handlers()
            self._loop_func()

        _LOGGER.info("GeckoUdpSocket thread finished")

    def _loop_func(self):
        # Opportunity for sub-class to get a thread loop
        pass

    def __repr__(self):
        return (
            f"{self.__class__.__name__} on {self._socket!r}\n"
            f"  receive_handlers={self._receive_handlers!r},\n"
            f"  send_handlers={self._send_handlers!r}\n"
            f"  isopen: {self.isopen}"
            f" isbusy: {self.isbusy}"
        )



""" Gecko automation support for sensors """

from .base import GeckoAutomationBase


class GeckoSensor(GeckoAutomationBase):
    """Sensors wrap accessors state with extra units and device
    class properties"""

    def __init__(self, facade, name, accessor, unit_accessor=None, device_class=None):
        super().__init__(facade, name, name.upper())
        self._accessor = accessor
        # Bubble up change notification
        accessor.watch(self._on_change)
        self._unit_of_measurement_accessor = unit_accessor
        if unit_accessor:
            unit_accessor.watch(self._on_change)
        self._device_class = device_class

    @property
    def state(self):
        """ The state of the sensor """
        return self._accessor.value

    @property
    def unit_of_measurement(self):
        """ The unit of measurement for the sensor, or None """
        if self._unit_of_measurement_accessor is None:
            return None
        return self._unit_of_measurement_accessor.value

    @property
    def device_class(self):
        """ The device class """
        return self._device_class

    @property
    def accessor(self):
        """ Access the accessor member """
        return self._accessor


########################################################################################
class GeckoBinarySensor(GeckoSensor):
    """ Binary sensors only have two states """

    @property
    def is_on(self):
        """ Determine if the sensor is on or not """
        state = self.state
        if isinstance(state, bool):
            return state
        return state != "OFF"
